#!/usr/bin/python -OO
"""
This is a geometric hash-table.

--stefan huber, 2007-12
"""


from rect import rect
import math


class geomhashtable:


		def __init__(self, rect, resx, resy):
				"""Initialize empty geometric hash-table with grid on rect
				with resolutions resx and resy in x and y direction, respectively"""

				assert( isinstance(resx,int))
				assert( isinstance(resy,int))
				assert( resx > 0 )
				assert( resy > 0 )
				assert( rect.left < rect.right )
				assert( rect.bot < rect.top )

				self.rect = rect
				self.resx = resx
				self.resy = resy
			
				#Initialize empty objects hash table
				self.objects = dict()

				#Empty geometric grid. We avoid . for performance
				table = dict()
				for j in xrange(0,resy):
					for i in xrange(0,resx):
						table[(i,j)] = []
				self.table = table



		def __repr__(self):
				txt = "[" + repr(self.rect) + " res " + repr((self.resx,self.resy)) + " on {"

				for j in xrange(0, self.resy):
					for i in xrange(0, self.resx):

							if len(self.table[i,j])>0:
								txt += repr( (i,j) ) + " : " + repr(self.table[i,j])
								txt += ", "	
				txt += "}]"
				return txt



		def __iter__(self):
				return iter(self.objects)

		def __contains__(self, obj):
				return obj in self.objects

		def __hash__(self):
				return hash( hash(self.rect) + hash(self.resx) + \
							hash(self.resy) + hash(self.objects))

		def __len__(self):
				return len(self.objects)
	
		def __delitem__(self, obj):

				if obj not in self.objects:
					return

				del self.objects[obj]

				#Add obj to all boxes intersected by obj
				for i,j in self.getIntersectingIterator(obj):
					self.table[i,j].remove(obj)


		def __iadd__(self, obj, eps=1e-11):

				if obj in self.objects:
					return self

				if not obj.containedInBox( self.rect, eps ):
					raise Exception(repr(obj) + " not contained in " + repr(self.rect))
				
				self.objects[obj] = obj
				
				#Add obj to all boxes intersected by obj
				for i,j in self.getIntersectingIterator(obj, eps):
					self.table[i,j].append(obj)

				return self


		def getHitObjects(self, obj, eps=0.0):
				"""Get a set of all items in hashtable which could intersect"""

				#Matching objects
				matched = set([])
	
				for i,j in self.getIntersectingIterator(obj, eps):
					submatch = set( self.table[i,j] )
					matched |= submatch

				return matched


		def getHitIndex(self, p, lb=True):
				"""Get index of grid-box hit by point p. If point lies on a boundary
				the most left/bot indices is taken if lb is set. Otherwise we take
				the most right/top one"""

				bb = self.rect

				#if not bb.containsPnt(p):
				#	raise Exception(repr(p) + " not contained in " + repr(self.rect))
							
				x,y = p
				i = (x - bb.left)*self.resx/(bb.right-bb.left)
				j = (y - bb.bot)*self.resy/(bb.top-bb.bot)

				i = max(0, min(i,self.resx))
				j = max(0, min(j,self.resy))

				if lb:
					return int(i), int(j)
				else:
					return int(math.ceil(i)), int(math.ceil(j))


		def rectOfGridIdx(self, i, j):
				"""Get the rectangle of the grid-box in the i-th column
				and in the j-th row"""

				assert( 0 <= i <= self.resx)
				assert( 0 <= j <= self.resy)

				deltax = float(self.rect.right-self.rect.left)/self.resx
				deltay = float(self.rect.top-self.rect.bot)/self.resy

				left = self.rect.left + i*deltax
				bot = self.rect.bot + j*deltay

				if i==self.resx-1:
					right = self.rect.right
				else:
					right = left+deltax

				if j==self.resy-1:
					top = self.rect.top
				else:
					top = bot+deltay

				return rect(left, right, bot, top)


		def getIntersectingIterator(self, obj, eps=0.0):
				"""Iterator enumerating all indices (i,j) such that
				the box with index (i,j) intersects with obj"""

				#Test every box if intersected by obj
				for i,j in self.getBBoxGridIterator(obj,eps):
					box = self.rectOfGridIdx(i,j)
					box.grow(eps)

					#Intersecting --> add to grid
					if obj.intersectsBox(box, 1e-11):
						yield i,j
				

		def getBBoxGridIterator(self, obj, eps=0.0):
				"""GridIterator for bounding box of obj"""

				#Get bouding box and indices for start and end
				bb = obj.boundingBox()
				bb.grow(eps)

				#Get indices
				si,sj = self.getHitIndex( bb.lb(), True)
				ei,ej = self.getHitIndex( bb.rt(), False)

				assert(si<=ei)
				assert(sj<=ej)

				for j in xrange(sj,ej+1):
					for i in xrange(si,ei+1):
						yield (i,j)

	
		def getGridIterator(self):
				"""Get iterator over all indices (i,j)"""
				for j in xrange(0, self.resy):
					for i in xrange(0, self.resx):
						yield (i,j)


		def addSplittingObj(ght, nobj, eps=0.0):
				"""Add obj and split obj and all elements in hash table by
				calling getSplittedBy"""

				#The parts of the object -- at the beginning the obj itself
				nparts = [nobj]
				hitobjs = ght.getHitObjects(nobj, eps)

				#Split nparts by every hitobj in hitobjs
				for hitobj in hitobjs:
						
					#Temporarie list
					npartstmp = []

					#Split every part of circ...
					for part in nparts:
						npartstmp += part.getSplittedBy(hitobj, eps)
					#The new list
					nparts = npartstmp


				#Now, split the hitobj in hitobjs by the new obj
				for hitobj in hitobjs:

					hitobjparts = hitobj.getSplittedBy(nobj, eps)

					if len(hitobjparts)>1 or hitobjparts[0]!=hitobj:
						del ght[hitobj]
						for part in hitobjparts:
							ght += part


				#Add the parts to geomhashtable
				for part in nparts:
					ght += part





	

