#!/usr/bin/python 
"""
This is a random circle generator

--stefan huber, 2007-12
"""

import sys

from geomhashtable import geomhashtable
from point import point
from rect import rect
from circle import circle
from arc import arc
from math import sqrt
from math import pi
from random import uniform
from time import clock
import math




def printUsage():
		print """Usage: %s --nmb <number-circles> [options]
  or: %s --help

With following options:
--gui                         Show gui
--maxrad <rel-max-size>=0.1   Max. radius relative to dimensio of grid
--resx <int>=100              Resolution of grid in x-direction
--resy <int>=100              Resolution of grid in y-direction
--cgal-file <filename>        Name of cgal dat-file
--vroni-file <filename>       Name of vroni sites-file

  
  """ % (sys.argv[0], sys.argv[0])


def parseParams():
		"""Parse CLI parameters"""

		nmb = None
		relMaxRad = 0.1
		gui = False
		resx = 100
		resy = 100
		cgalfn = None
		vronifn = None

		try:

				#Parse all tokens
				p=1
				while p < len(sys.argv):

					tok = sys.argv[p]

					if tok=="--nmb":
						p += 1
						nmb = int(sys.argv[p])

					elif tok=="--resx":
						p += 1
						resx = int(sys.argv[p])

					elif tok=="--resy":
						p += 1
						resy = int(sys.argv[p])

					elif tok=="--maxrad":
						p += 1
						relMaxRad = float(sys.argv[p])

					elif tok=="--cgal-file":
						p += 1
						cgalfn = sys.argv[p]

					elif tok=="--vroni-file":
						p += 1
						vronifn = sys.argv[p]

					elif tok=="--gui":
						gui = True

					elif tok=="--help":
						printUsage()
						sys.exit(0)

					else:
						print "Unkown option"
						printUsage()
						sys.exit(-1)

					#Next token
					p += 1

		except ValueError, e:
			print "Error parsing argument: " + str(e)
			printUsage()
			sys.exit(-1)

		if nmb==None or nmb<=0:
			print "Please give positive number of circles."
			printUsage()
			sys.exit(-2)

		if not 0 < relMaxRad <= 0.5:
			print "Relative radius can not be bigger than 0.5 or less-or-equal than 0."
			printUsage()
			sys.exit(-3)

		if resx<=0 or resy<=0:
			print "Grid resolution must be a positive integer."
			printUsage()
			sys.exit(-5)


		return nmb,relMaxRad,gui,resx,resy,cgalfn,vronifn




def compRandCircles( nmb, relMaxRad, resx, resy, verbose=False):

		assert(0 <= relMaxRad <= 1)

		def randCircle():
			rad = uniform(0.05*relMaxRad,relMaxRad)
			return circle( (uniform(rad,1-rad), uniform(rad,1-rad)), rad )


		#define the bounding box and the hash table
		bb = rect(0, 1, 0, 1)
		ght = geomhashtable( bb, resx, resy )

		if verbose:
			print "created hash-table after %f sec" % (clock(),)

		#Create the random circles
		randcircles = [ randCircle() for i in xrange(0, nmb) ]

		if verbose:
			print "created random circles after %f sec" % (clock(),)

		#Insert those circles...
		i=0
		for circ in randcircles:

			i += 1
			ght.addSplittingObj(circ, 1e-10)
					
			#Some status report
			if i%(max(1,nmb/20)) == 0 and i>0:
					now = clock()
					remaining = now*(nmb-i)/i

					if verbose:
						print "\tinserted %d-th circle...   (elapsed %f sec, remaining %f sec)" % \
								(i, now, remaining )

		if verbose:
			print "Finished after %f sec." % (clock(),)
			print "Hashtable contains", len(ght), "objects."

		return randcircles, ght


def extractArcs(ght):
		arcs = []

		for obj in ght:

			if isinstance(obj, arc):
				arcs += [obj]

			elif isinstance(obj, circle):
				x,y,r = obj.x, obj.y, obj.radius
				arcs += [ arc((x, y), r, 0, pi/2), \
							arc((x, y), r, pi/2, pi/2), \
							arc((x, y), r, pi, pi/2), \
							arc((x, y), r, 3*pi/2, pi/2)]

		return arcs


def writeCgalFile(circles, fn):

		file = open(fn, "w")

		i = 0
		for circ in circles:

			i += 1
			if i%1000==0 and i>0:
				print("Writing circle #%d" % (i,))

			assert( isinstance(circ, circle))
			file.write("%.17f %.17f %.17f\n" % (circ.x, circ.y, circ.radius))

		file.close()


def writeVroniFile(arcs, fn):

		bb = rect(0, 1, 0, 1)
		ght = geomhashtable( bb, 100, 100 )
		file = open(fn, "w")

		i=0
		for a in arcs:

			i += 1
			if i%1000==0 and i>0:
				print("Writing arc #%d" % (i,))

			assert( isinstance(a, arc))

			sx,sy = a.start()
			ex,ey = a.end()


			#BUG -- implement ght for this purpose!
			for pnt in ght.getHitObjects( point((sx,sy)), 1e-13**2 ):
				px, py = pnt.coord()
				if (sx-px)**2 + (sy-py)**2 < 1e-14**2:
					sx,sy = px,py
					break
			else:
				ght += point((sx,sy))

			for pnt in ght.getHitObjects( point((ex,ey)), 1e-13**2 ):
				px, py = pnt.coord()
				if (ex-px)**2 + (ey-py)**2 < 1e-14**2:
					ex,ey = px,py
					break
			else:
				ght += point((ex,ey))

			file.write("1 %.17f %.17f %.17f %.17f %.17f %.17f\n" % (sx, sy, ex, ey, a.x, a.y))

		file.close()


if __name__ == "__main__":

		try:
				import psyco
				psyco.full()
		except ImportError:
				print "Error importing psyco."
				pass

		clock()
		nmb,relMaxRad,gui,resx,resy,cgalfn,vronifn = parseParams();


		#Compute random circles
		randcircles, ght = compRandCircles(nmb, relMaxRad, resx, resy, True)


		if cgalfn!=None:
			writeCgalFile(randcircles, cgalfn)
	
		if vronifn!=None:
			writeVroniFile(extractArcs(ght), vronifn)


		#User gave the --gui flag 
		if gui:

				try:
					from PyQt4 import QtGui, QtCore
					import qgeomhashtablewidget

					app = QtGui.QApplication(sys.argv)
					win = qgeomhashtablewidget.MainWindow(ght, 512)
					win.show()

					app.connect(app, QtCore.SIGNAL('lastWindowClosed()'), app, QtCore.SLOT('quit()') )
					sys.exit(app.exec_());

				except (ImportError):
					print "Unable to import PyQt4."
					sys.exit(-4)


					



